# cython: language_level=3
#  Drakkar-Software OctoBot-Commons
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


cpdef object parse_symbol(str symbol)   # return object to allow exception propagation
cpdef str merge_symbol(str symbol)
cpdef str merge_currencies(
        str currency,
        str market,
        str settlement_asset=*,
        str market_separator=*,
        str settlement_separator=*
)
cpdef str convert_symbol(
    str symbol,
    str symbol_separator,
    str new_symbol_separator=*,
    bint should_uppercase=*,
    bint should_lowercase=*,
    bint base_and_quote_only=*
)
